"""Utility functions for distributions
"""
import numpy as np

MIN_EXPONENT = np.log(np.float32(1e-32))
MAX_EXPONENT = np.log(np.finfo("float32").max) - 1


def safe_exp(x):
    """Like np.exp, but clipped to prevent overflow (in float32 world)

    Ensures that
        1. large numbers cannot hit infinity
        2. small numbers cannot hit precisely zero

    NB: The limits are chosen such that we have some stability in subsequent
    computations. E.g the minimum returned value should be safe in a division
    with a numerator of size up to ~1e6.
    """
    # TODO: Empirically refine these limits using different datasets.
    # Can we clip more without losing accuracy?
    x_clipped = np.clip(x, a_min=MIN_EXPONENT, a_max=MAX_EXPONENT)
    return np.exp(x_clipped)


def check_all_integer(x):
    if not all(x == x.astype(int)):
        raise ValueError("All values of target must be integers!")


def check_all_ge_zero(x):
    if not all(x >= 0):
        raise ValueError("All values of target must be >=0!")


def check_all_gt_zero(x):
    if not all(x > 0):
        raise ValueError("All values of target must be > 0!")



from functools import singledispatch

import numpy as np


@singledispatch
def to_serializable(val):
    return str(val)


@to_serializable.register
def _(val: np.float32):
    return np.float64(val)



"""Test suite for XGBDistribution model"""

import os
import pickle

import pytest

import joblib
import numpy as np
from sklearn.exceptions import NotFittedError

from xgboost_distribution.distributions import AVAILABLE_DISTRIBUTIONS
from xgboost_distribution.model import XGBDistribution


def test_XGBDistribution_early_stopping_fit(small_train_test_data):
    """Integration test to ensure end-to-end functionality"""

    X_train, X_test, y_train, y_test = small_train_test_data
    model = XGBDistribution(
        distribution="normal", max_depth=3, n_estimators=500, early_stopping_rounds=10
    )
    model.fit(X_train, y_train, eval_set=[(X_test, y_test)])
    evals_result = model.evals_result()

    assert model.best_iteration == 5
    assert isinstance(evals_result, dict)


def test_XGBDistribution_early_stopping_fit_single_param_distribution(
    small_train_test_count_data,
):
    """Integration test for single param dist (which operate on squeezed arrays)"""

    X_train, X_test, y_train, y_test = small_train_test_count_data

    model = XGBDistribution(
        distribution="exponential",
        max_depth=3,
        n_estimators=500,
        early_stopping_rounds=10,
    )
    model.fit(X_train, y_train, eval_set=[(X_test, y_test)])
    evals_result = model.evals_result()

    assert model.best_iteration == 10
    assert isinstance(evals_result, dict)


def test_XGBDistribution_early_stopping_predict(small_train_test_data):
    """Check that predict with early stopping uses correct ntrees"""
    X_train, X_test, y_train, y_test = small_train_test_data

    model = XGBDistribution(
        distribution="normal", max_depth=3, n_estimators=500, early_stopping_rounds=10
    )
    model.fit(X_train, y_train, eval_set=[(X_test, y_test)])

    mean, var = model.predict(X_test)
    mean_iter, var_iter = model.predict(
        X_test, iteration_range=(0, model.best_iteration + 1)
    )
    np.testing.assert_array_equal(mean, mean_iter)
    np.testing.assert_array_equal(var, var_iter)


def test_objective_overwrite_by_repeated_fit(small_X_y_data):
    """In the fit step, we set the objective to distribution:normal (e.g.)

    This needs to be re-set at each fit, as it's not a standard xgboost objective.
    """
    X, y = small_X_y_data
    model = XGBDistribution()
    model.fit(X, y)
    model.fit(X, y)


def test_distribution_set_param(small_X_y_data):
    """Check that updating the distribution params works"""
    X, y = small_X_y_data

    model = XGBDistribution(distribution="abnormal")
    with pytest.raises(ValueError):
        model.fit(X, y)  # fails as distribution does not exist

    model.set_params(**{"distribution": "normal"})
    model.fit(X, y)


@pytest.mark.parametrize("distribution", ["normal"])
def test_fit_with_sample_weights(small_X_y_data, distribution):
    X, y = small_X_y_data

    random_weights = np.random.choice([1, 2], len(X))
    model = XGBDistribution(distribution=distribution, n_estimators=2)
    preds_without_weights = model.fit(X, y).predict(X)
    preds_with_weights = model.fit(X, y, sample_weight=random_weights).predict(X)

    with pytest.raises(AssertionError):
        np.testing.assert_array_equal(preds_without_weights.loc, preds_with_weights.loc)
        np.testing.assert_array_equal(
            preds_without_weights.scale, preds_with_weights.scale
        )


def test_sample_weights_eval_set(small_train_test_data):
    """Check weights for eval sets change NLL during training"""
    X_train, X_test, y_train, y_test = small_train_test_data

    weights_train = np.random.choice([1, 2], len(X_train))
    weights_test = np.random.choice([1, 2], len(X_test))

    model = XGBDistribution(distribution="normal", n_estimators=2)
    model.fit(
        X_train, y_train, sample_weight=weights_train, eval_set=[(X_test, y_test)]
    )
    evals_result_without_weights = model.evals_result()
    nll_without_weights = evals_result_without_weights["validation_0"][
        "NormalDistribution-NLL"
    ]

    model.fit(
        X_train,
        y_train,
        sample_weight=weights_train,
        eval_set=[(X_test, y_test)],
        sample_weight_eval_set=[weights_test],
    )
    evals_result_with_weights = model.evals_result()
    nll_with_weights = evals_result_with_weights["validation_0"][
        "NormalDistribution-NLL"
    ]

    assert all(nll_with_weights[i] != nll_without_weights[i] for i in range(2))


# -------------------------------------------------------------------------------------
# Failure modes
# -------------------------------------------------------------------------------------


def test_fit_1d_X_fit_fails(small_X_y_data):
    X, y = small_X_y_data
    X = X[..., 0]  # single 1d features

    model = XGBDistribution()
    with pytest.raises(ValueError):
        model.fit(X, y)


def test_predict_before_fit_fails(small_X_y_data):
    X, y = small_X_y_data

    model = XGBDistribution()
    with pytest.raises(NotFittedError):
        model.predict(X)


def test_setting_objective_in_init_fails():
    with pytest.raises(ValueError):
        XGBDistribution(objective="binary:logistic")


# -------------------------------------------------------------------------------------
#  Internal tests
# -------------------------------------------------------------------------------------


@pytest.mark.parametrize(
    "distribution, expected_margin",
    [
        ("normal", np.array([[0.5, np.log(0.5)], [0.5, np.log(0.5)]])),
        ("poisson", np.array([[np.log(0.5)], [np.log(0.5)]])),
    ],
)
def test_get_base_margin(distribution, expected_margin):
    """Test that base_margin are created as expected for each sample"""
    model = XGBDistribution(distribution=distribution)
    X = np.array([[1], [1]])
    y = np.array([1, 0])
    model.fit(X, y)

    margin = model._get_base_margin(n_samples=2)
    np.testing.assert_array_equal(margin, expected_margin)


@pytest.mark.parametrize(
    "distribution",
    list(AVAILABLE_DISTRIBUTIONS.keys()),
)
def test_objective_and_evaluation_funcs_callable(distribution):
    model = XGBDistribution(distribution=distribution)
    assert callable(model._objective_func())
    assert callable(model._evaluation_func())


# -------------------------------------------------------------------------------------
#  IO tests
# -------------------------------------------------------------------------------------


def assert_model_equivalence(model_a, model_b, X):
    np.testing.assert_almost_equal(
        model_a._starting_params, model_b._starting_params, decimal=9
    )
    assert model_a._distribution.__class__ == model_b._distribution.__class__

    preds_a = model_a.predict(X)
    preds_b = model_b.predict(X)

    for param_a, param_b in zip(preds_a, preds_b):
        np.testing.assert_array_equal(param_a, param_b)


@pytest.mark.parametrize(
    "model_format",
    ["json", "ubj"],
)
def test_XGBDistribution_save_and_load_model(small_X_y_data, model_format, tmpdir):
    X, y = small_X_y_data
    model = XGBDistribution(n_estimators=10, distribution="laplace")
    model.fit(X, y)

    model_path = os.path.join(tmpdir, f"model.{model_format}")
    model.save_model(model_path)

    saved_model = XGBDistribution()
    saved_model.load_model(model_path)

    assert_model_equivalence(model_a=model, model_b=saved_model, X=X)


def test_XGBDistribution_pickle_dump_and_load(small_X_y_data, tmpdir):
    X, y = small_X_y_data
    model = XGBDistribution(n_estimators=10)
    model.fit(X, y)

    model_path = os.path.join(tmpdir, "model.pkl")

    with open(model_path, "wb") as fd:
        pickle.dump(model, fd)

    with open(model_path, "rb") as fd:
        saved_model = pickle.load(fd)

    assert_model_equivalence(model_a=model, model_b=saved_model, X=X)


class XGBDistributionWrapper:
    def __init__(self, params):
        self.model1 = XGBDistribution(**params)
        self.model2 = XGBDistribution(**params)

    def fit(self, X, y):
        self.model1.fit(X, y)
        self.model2.fit(X, y)
        return self


def test_XGBDistribution_wrapper_joblib_dump_and_load(small_X_y_data, tmpdir):
    X, y = small_X_y_data
    model = XGBDistributionWrapper({"n_estimators": 2})
    model.fit(X, y)

    model_path = os.path.join(tmpdir, "model.pkl")

    with open(model_path, "wb") as fd:
        joblib.dump(model, fd)

    with open(model_path, "rb") as fd:
        saved_model = joblib.load(fd)

    assert_model_equivalence(model_a=model.model1, model_b=saved_model.model1, X=X)
    assert_model_equivalence(model_a=model.model2, model_b=saved_model.model2, X=X)



"""LogNormal distribution"""

from collections import namedtuple

import numpy as np
from scipy.stats import lognorm

from xgboost_distribution.compat import linalg_solve
from xgboost_distribution.distributions.base import BaseDistribution
from xgboost_distribution.distributions.utils import (
    MAX_EXPONENT,
    MIN_EXPONENT,
    check_all_gt_zero,
)

MIN_LOG_SCALE = MIN_EXPONENT / 2
MAX_LOG_SCALE = MAX_EXPONENT / 2

Params = namedtuple("Params", ("scale", "s"))


class LogNormal(BaseDistribution):
    """LogNormal distribution with log scoring.

    Definition:

        f(x) = exp( -[ (log(x) - log(scale)) / (2 s^2) ]^2 / 2 ) / s


    with parameters (scale, s).

    We reparameterize:
        s -> log(s) = a
        scale -> log(scale) = b

    Note that b essentially becomes the 'loc' of the distribution:

        log(x/scale) / s = ( log(x) - log(scale) ) / s

    which can then be taken analogous to the normal distribution's

        (x - loc) / scale

    Hence we can re-use the computations in `distribution.normal`, exchanging:

        y -> log(y)
        scale -> s

    """

    @property
    def params(self):
        return Params._fields

    def check_target(self, y):
        check_all_gt_zero(y)

    def gradient_and_hessian(self, y, params, natural_gradient=True):
        """Gradient and diagonal hessian"""

        log_y = np.log(y)
        loc, log_s = self._safe_params(params)  # note loc = log(scale)
        var = np.exp(2 * log_s)

        grad = np.zeros(shape=(len(y), 2), dtype="float32")
        grad[:, 0] = (loc - log_y) / var
        grad[:, 1] = 1 - ((loc - log_y) ** 2) / var

        if natural_gradient:
            fisher_matrix = np.zeros(shape=(len(y), 2, 2), dtype="float32")
            fisher_matrix[:, 0, 0] = 1 / var
            fisher_matrix[:, 1, 1] = 2

            grad = linalg_solve(fisher_matrix, grad)
            hess = np.ones(shape=(len(y), 2), dtype="float32")  # constant hessian
        else:
            hess = np.zeros(shape=(len(y), 2), dtype="float32")
            hess[:, 0] = 1 / var
            hess[:, 1] = 2 * ((log_y - loc) ** 2) / var

        return grad, hess

    def loss(self, y, params):
        scale, s = self.predict(params)
        return "LogNormal-NLL", -lognorm.logpdf(y, s=s, scale=scale)

    def predict(self, params):
        log_scale, log_s = self._safe_params(params)
        scale, s = np.exp(log_scale), np.exp(log_s)
        return Params(scale=scale, s=s)

    def starting_params(self, y):
        log_y = np.log(y)
        return Params(scale=np.mean(log_y), s=np.log(np.std(log_y)))

    def _safe_params(self, params):
        """Return safe log_scale (loc) and log_s from params"""
        loc = params[:, 0]
        log_s = np.clip(params[:, 1], a_min=MIN_LOG_SCALE, a_max=MAX_LOG_SCALE)
        return loc, log_s



"""Metrics for evaluating xgboost-distribution scores
"""
from typing import Callable, Tuple

import numpy as np
import scipy
from xgboost._typing import ArrayLike


def get_ll_score_func(
    distribution: str,
) -> Callable[[ArrayLike, Tuple[np.ndarray]], float]:
    """Get log-likelihood scoring function for a given distribution

    Parameters
    ----------
    distribution : str

    Returns
    -------
    Callable
        Scoring function
    """
    dist_ll = {
        "exponential": scipy.stats.expon.logpdf,
        "laplace": scipy.stats.laplace.logpdf,
        "log-normal": scipy.stats.lognorm.logpdf,
        "negative-binomial": scipy.stats.nbinom.logpmf,
        "normal": scipy.stats.norm.logpdf,
        "poisson": scipy.stats.poisson.logpmf,
    }

    def score_func(y, y_pred):
        return dist_ll[distribution](y, *y_pred).mean()

    return score_func



"""Test the base distribution functionality"""

import pytest

import numpy as np

from xgboost_distribution.distributions import (
    AVAILABLE_DISTRIBUTIONS,
    Normal,
    format_distribution_name,
    get_distribution,
    get_distribution_doc,
)


def test_get_distribution():
    name = "normal"
    normal_dist = get_distribution(name)
    assert isinstance(normal_dist, Normal)


def test_get_distribution_raises():
    name = "i-do-not-exist"
    with pytest.raises(ValueError):
        get_distribution(name)


def test_get_distribution_doc():
    distribution_doc = get_distribution_doc()
    assert isinstance(distribution_doc, str)


@pytest.mark.parametrize(
    "name, expected_name",
    [("LogNormal", "log-normal"), ("NegativeBinomial", "negative-binomial")],
)
def test_format_distribution_name(name, expected_name):
    formatted_name = format_distribution_name(name)
    assert formatted_name == expected_name


@pytest.mark.parametrize("distribution_name", AVAILABLE_DISTRIBUTIONS.keys())
def test_distribution_starting_params_shape(distribution_name):
    """We need to get as many starting params as distribution params"""
    y = np.random.choice([1, 2], 10)

    distribution = get_distribution(distribution_name)
    starting_params = distribution.starting_params(y=y)

    assert len(starting_params) == len(distribution.params)


@pytest.mark.parametrize("distribution_name", AVAILABLE_DISTRIBUTIONS.keys())
def test_distribution_loss_shape(distribution_name):
    """Ensure that evaluation fns return expect nd.array shape"""
    y = np.random.choice([1, 2], 10)

    distribution = get_distribution(distribution_name)
    starting_params = distribution.starting_params(y=y)

    params = np.concatenate([starting_params]) * np.ones_like(y).reshape(-1, 1)

    if len(distribution.params) == 1:
        params = params.squeeze()  # for 1 param we get a squeezed array (n,) from xgb

    loss_name, loss = distribution.loss(y, params)

    assert isinstance(loss_name, str)
    assert loss.shape == y.shape



import pytest

import numpy as np

from xgboost_distribution.distributions import Normal


@pytest.fixture
def normal():
    return Normal()


@pytest.mark.parametrize(
    "y, params, natural_gradient, expected_grad",
    [
        (
            np.array([0, 0]),
            np.array([[0, 1], [1, 0]]),
            True,
            np.array([[0, 0.5], [1, 0]]),
        ),
        (
            np.array([0, 0]),
            np.array([[0, 1], [1, 0]]),
            False,
            np.array([[0, 1], [1, 0]]),
        ),
    ],
)
def test_gradient_calculation(normal, y, params, natural_gradient, expected_grad):
    grad, _ = normal.gradient_and_hessian(y, params, natural_gradient=natural_gradient)
    np.testing.assert_array_equal(grad, expected_grad)


@pytest.mark.parametrize(
    "y, params",
    [
        (
            np.array([34.7, 20.1], dtype="float32"),
            np.array([[31.2, 50], [11, 6.7]], dtype="float32"),
        ),
        (
            np.array([34.7, 20.1], dtype="float32"),
            np.array([[1.2e3, -100], [11, 6.7]], dtype="float32"),
        ),
    ],
)
def test_overflow_stability(normal, y, params):
    """Test stability against large/small values produced by xgboost"""

    grad, _ = normal.gradient_and_hessian(y, params)
    assert isinstance(grad, np.ndarray)

    # Var should not have inf or zeros
    _, scale = normal.predict(params)
    var = scale**2
    assert all(np.isfinite(var))
    assert np.all(var)



"""Distribution base class"""

from abc import ABC, abstractmethod
from typing import Tuple

import numpy as np


class BaseDistribution(ABC):
    """Base class distribution for XGBDistribution.

    Note that distributions are stateless, hence a distribution is just a collection of
    functions that operate on the data (`y`) and the outputs of the xgboost (`params`).
    """

    def check_target(self, y) -> None:
        """Ensure that the target is compatible with the chosen distribution"""

    @property
    @abstractmethod
    def params(self) -> Tuple[str, ...]:
        """The parameter names of the distribution"""

    @abstractmethod
    def starting_params(self, y) -> Tuple[float, ...]:
        """The starting parameters of the distribution"""

    @abstractmethod
    def gradient_and_hessian(
        self, y, params, natural_gradient=True
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Compute the gradient and hessian of the distribution"""

    @abstractmethod
    def loss(self, y, params) -> Tuple[str, np.ndarray]:
        """Evaluate the per sample loss (typically negative log-likelihood)"""

    @abstractmethod
    def predict(self, params) -> Tuple[np.ndarray, ...]:
        """Predict the parameters of a given distribution"""
